# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 OpenStack LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""SQL backends for the various services."""
import functools
import re
import time

import sqlalchemy as sql
import sqlalchemy.engine.url
from sqlalchemy.exc import DisconnectionError
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext import declarative
import sqlalchemy.orm
import sqlalchemy.pool
from sqlalchemy import types as sql_types
from sqlalchemy.orm.attributes import InstrumentedAttribute

from keystone.common import logging
from keystone import config
from keystone import exception
from keystone.openstack.common import jsonutils

from keystone.exception import DBError

CONF = config.CONF
LOG = logging.getLogger(__name__)

# maintain a single engine reference for sqlite in-memory
GLOBAL_ENGINE = None


ModelBase = declarative.declarative_base()


# For exporting to other modules
Column = sql.Column
String = sql.String
ForeignKey = sql.ForeignKey
DateTime = sql.DateTime
IntegrityError = sql.exc.IntegrityError
NotFound = sql.orm.exc.NoResultFound
Boolean = sql.Boolean
Text = sql.Text
UniqueConstraint = sql.UniqueConstraint


def initialize_decorator(init):
    """Ensure that the length of string field do not exceed the limit.

    This decorator check the initialize arguments, to make sure the
    length of string field do not exceed the length limit, or raise a
    'StringLengthExceeded' exception.

    Use decorator instead of inheritance, because the metaclass will
    check the __tablename__, primary key columns, etc. at the class
    definition.

    """
    def initialize(self, *args, **kwargs):
        cls = type(self)
        for k, v in kwargs.items():
            if hasattr(cls, k):
                attr = getattr(cls, k)
                if isinstance(attr, InstrumentedAttribute):
                    column = attr.property.columns[0]
                    if isinstance(column.type, String):
                        if column.type.length and \
                                column.type.length < len(str(v)):
                            #if signing.token_format == 'PKI', the id will
                            #store it's public key which is very long.
                            if config.CONF.signing.token_format == 'PKI' and \
                                    self.__tablename__ == 'token' and \
                                    k == 'id':
                                continue

                            raise exception.StringLengthExceeded(
                                string=v, type=k, length=column.type.length)

        init(self, *args, **kwargs)
    return initialize

ModelBase.__init__ = initialize_decorator(ModelBase.__init__)


def set_global_engine(engine):
    global GLOBAL_ENGINE
    GLOBAL_ENGINE = engine


def get_global_engine():
    return GLOBAL_ENGINE


def is_db_connection_error(args):
    """Return True if error in connecting to db."""
    # NOTE(adam_g): This is currently MySQL specific and needs to be extended
    #               to support Postgres and others.
    conn_err_codes = ('2002', '2003', '2006', '2013', '1213')
    for err_code in conn_err_codes:
        if str(args).find(err_code) != -1:
            return True
    return False


# Special Fields
class JsonBlob(sql_types.TypeDecorator):

    impl = sql.Text

    def process_bind_param(self, value, dialect):
        return jsonutils.dumps(value)

    def process_result_value(self, value, dialect):
        return jsonutils.loads(value)


class DictBase(object):
    attributes = []

    @classmethod
    def from_dict(cls, d):
        new_d = d.copy()

        new_d['extra'] = dict((k, new_d.pop(k)) for k in d.iterkeys()
                              if k not in cls.attributes and k != 'extra')

        return cls(**new_d)

    def to_dict(self, include_extra_dict=False):
        """Returns the model's attributes as a dictionary.

        If include_extra_dict is True, 'extra' attributes are literally
        included in the resulting dictionary twice, for backwards-compatibility
        with a broken implementation.

        """
        d = self.extra.copy()
        for attr in self.__class__.attributes:
            d[attr] = getattr(self, attr)

        if include_extra_dict:
            d['extra'] = self.extra.copy()

        return d

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __iter__(self):
        self._i = iter(sqlalchemy.orm.object_mapper(self).columns)
        return self

    def next(self):
        n = self._i.next().name
        return n

    def update(self, values):
        """Make the model object behave like a dict."""
        for k, v in values.iteritems():
            setattr(self, k, v)

    def iteritems(self):
        """Make the model object behave like a dict.

        Includes attributes from joins.

        """
        return dict([(k, getattr(self, k)) for k in self])
        #local = dict(self)
        #joined = dict([(k, v) for k, v in self.__dict__.iteritems()
        #               if not k[0] == '_'])
        #local.update(joined)
        #return local.iteritems()


class MySQLPingListener(object):

    """
    Ensures that MySQL connections checked out of the
    pool are alive.

    Borrowed from:
    http://groups.google.com/group/sqlalchemy/msg/a4ce563d802c929f

    Error codes caught:
    * 2006 MySQL server has gone away
    * 2013 Lost connection to MySQL server during query
    * 2014 Commands out of sync; you can't run this command now
    * 2045 Can't open shared memory; no answer from server (%lu)
    * 2055 Lost connection to MySQL server at '%s', system error: %d

    from http://dev.mysql.com/doc/refman/5.6/en/error-messages-client.html
    """

    def checkout(self, dbapi_con, con_record, con_proxy):
        try:
            dbapi_con.cursor().execute('select 1')
        except dbapi_con.OperationalError as e:
            if e.args[0] in (2006, 2013, 2014, 2045, 2055):
                logging.warn(_('Got mysql server has gone away: %s'), e)
                raise DisconnectionError("Database server went away")
            else:
                raise


# Backends
class Base(object):
    _engine = None
    _sessionmaker = None

    def get_session(self, autocommit=True, expire_on_commit=False):
        """Return a SQLAlchemy session."""
        self._engine = self._engine or self.get_engine()
        self._sessionmaker = self._sessionmaker or self.get_sessionmaker(
            self._engine)
        session = self._sessionmaker()
        session.expire_all()
        return session

    def get_engine(self, allow_global_engine=True):
        """Return a SQLAlchemy engine.

        If allow_global_engine is True and an in-memory sqlite connection
        string is provided by CONF, all backends will share a global sqlalchemy
        engine.

        """
        def new_engine():
            connection_dict = sql.engine.url.make_url(CONF.sql.connection)

            engine_config = {
                'convert_unicode': True,
                'echo': CONF.debug and CONF.verbose,
                'pool_recycle': CONF.sql.idle_timeout,
            }

            if 'sqlite' in connection_dict.drivername:
                engine_config['poolclass'] = sqlalchemy.pool.StaticPool
            elif 'mysql' in connection_dict.drivername:
                engine_config['listeners'] = [MySQLPingListener()]

            engine = sql.create_engine(CONF.sql.connection, **engine_config)

            try:
                engine.connect()
            except OperationalError, e:
                if not is_db_connection_error(e.args[0]):
                    raise

                next_interval = CONF.sql.retry_interval
                remaining = CONF.sql.max_retries
                if remaining == -1:
                    remaining = 'infinite'
                while True:
                    msg = 'SQL connection failed. %s attempts left.'
                    logging.warn(msg % remaining)
                    if remaining != 'infinite':
                        remaining -= 1
                    time.sleep(next_interval)
                    try:
                        engine.connect()
                        break
                    except OperationalError, e:
                        if ((remaining != 'infinite' and remaining == 0) or
                                not is_db_connection_error(e.args[0])):
                            raise
                        if CONF.sql.inc_retry_interval:
                            next_interval = min(next_interval * 2,
                                                CONF.sql.max_retry_interval)

            return engine

        engine = get_global_engine() or new_engine()

        # auto-build the db to support wsgi server w/ in-memory backend
        if allow_global_engine and CONF.sql.connection == 'sqlite://':
            ModelBase.metadata.create_all(bind=engine)

        if allow_global_engine:
            set_global_engine(engine)

        return engine

    def wrap_db_error(self, f):
        """Function wrapper to capture DB errors

        If an exception is thrown by the wrapped function,
        determine if it represents a database connection error.
        If so, retry the wrapped function, and repeat until it succeeds
        or we reach a configurable maximum number of retries.
        If it is not a connection error, or we exceeded the retry limit,
        raise a DBError.
        """
        def _wrap_db_error(*args, **kwargs):

            connection_dict = sql.engine.url.make_url(CONF.sql.connection)
            if 'mysql' in connection_dict.drivername:
                import MySQLdb
                operational_errors = (MySQLdb.OperationalError,
                                      OperationalError)
            else:
                operational_errors = (OperationalError,)

            next_interval = CONF.sql.retry_interval
            remaining = CONF.sql.max_retries
            if remaining == -1:
                remaining = 'infinite'
            while True:
                try:
                    return f(*args, **kwargs)
                except operational_errors as e:
                    if is_db_connection_error(e.args[0]):
                        if remaining == 0:
                            logging.warn('DB exceeded retry limit.')
                            raise DBError(e)
                        if remaining != 'infinite':
                            remaining -= 1
                        logging.warn('DB connection error: %s\n'
                                     'retrying in %i seconds.'
                                     % (e.args[0], next_interval))
                        time.sleep(next_interval)
                        if CONF.sql.inc_retry_interval:
                            next_interval = min(next_interval * 2,
                                                CONF.sql.max_retry_interval)
                    else:
                        logging.warn('DB exception wrapped.'
                                     'Will not retry because'
                                     'it is not connection'
                                     'error: %s.' % (e.args[0]))
                        raise DBError(e)
                except InvalidRequestError as e:
                    exc_re = re.search('(.*)Original\s+exception\s+was:'
                                       '\s+\((.*)\)\s+\((.*)\)', e.args[0])
                    if exc_re:
                        invalidreqerror_msg = exc_re.group(1)
                        original_exception_type = exc_re.group(2)
                        original_exception_value = exc_re.group(3)
                        if re.search('rollback', invalidreqerror_msg):
                            self.get_session().rollback()
                        if is_db_connection_error(original_exception_value):
                            if remaining == 0:
                                LOG.exception(_('DB exceeded retry limit.'))
                                raise DBError(e)
                            if remaining != 'infinite':
                                remaining -= 1
                            LOG.exception(_('DB connection error, '
                                            'retrying in %i seconds.')
                                          % next_interval)
                            time.sleep(next_interval)
                            if CONF.sql.inc_retry_interval:
                                next_interval = min(
                                    next_interval * 2,
                                    CONF.sql.max_retry_interval)
                        else:
                            LOG.exception(_('DB exception wrapped.'))
                            raise DBError(e)
                    else:
                        LOG.exception(_('DB exception wrapped.'))
                        raise DBError(e)
                except Exception, e:
                    logging.warn('DB exception: %s.' % (e.args[0]))
                    raise e

        _wrap_db_error.func_name = f.func_name
        return _wrap_db_error

    def get_sessionmaker(self, engine, autocommit=True,
                         expire_on_commit=False):
        """Return a SQLAlchemy sessionmaker using the given engine."""
        class OurQuery(sqlalchemy.orm.query.Query):
            pass

        query = OurQuery
        query.all = self.wrap_db_error(query.all)
        query.first = self.wrap_db_error(query.first)
        query.delete = self.wrap_db_error(query.delete)
        query.update = self.wrap_db_error(query.update)
        session = sqlalchemy.orm.sessionmaker(
            bind=engine,
            autocommit=autocommit,
            expire_on_commit=expire_on_commit,
            query_cls=query)
        session.begin = self.wrap_db_error(session.begin)
        session.execute = self.wrap_db_error(session.execute)
        session.flush = self.wrap_db_error(session.flush)
        session.query = self.wrap_db_error(session.query)
        session.delete = self.wrap_db_error(session.delete)
        session.rollback = self.wrap_db_error(session.rollback)
        session.commit = self.wrap_db_error(session.commit)
        return session


def handle_conflicts(type='object'):
    """Converts IntegrityError into HTTP 409 Conflict."""
    def decorator(method):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except IntegrityError as e:
                raise exception.Conflict(type=type, details=str(e.orig))
        return wrapper
    return decorator
